#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.Exceptions import InvalidInputException

class KwargsChecker():
	def __init__(self, required_arguments = None, optional_arguments = None, check_functions = None):
		assert((required_arguments is None) or isinstance(required_arguments, set))
		assert((optional_arguments is None) or isinstance(optional_arguments, set))
		self._required_arguments = required_arguments or set()
		self._optional_arguments = optional_arguments or set()
		self._allowed_arguments = self._required_arguments | self._optional_arguments
		self._check_functions = check_functions

	def is_optional_argument(self, name):
		return name in self._optional_arguments

	def check_single(self, arg, hint = None):
		if arg not in self._allowed_arguments:
			if hint is None:
				raise InvalidInputException("%s is not a valid argument." % (arg))
			else:
				raise InvalidInputException("%s is not a valid argument for %s." % (arg, hint))

	def check(self, args, hint = None):
		args = set(args)
		unknown_arguments = args - self._allowed_arguments
		missing_arguments = self._required_arguments - args

		errors = [ ]
		if len(unknown_arguments) > 0:
			if len(unknown_arguments) == 1:
				errors.append("unknown argument: %s" % (list(unknown_arguments)[0]))
			else:
				errors.append("%d unknown arguments: %s" % (len(unknown_arguments), ", ".join(sorted(unknown_arguments))))
		if len(missing_arguments) > 0:
			errors.append("required argument(s) missing: %s" % (", ".join(sorted(missing_arguments))))

		if len(errors) > 0:
			if len(errors) == 1:
				msg = "There was an error with the arguments"
			else:
				msg = "There were %d error(s) with the arguments" % (len(errors))
			if hint is not None:
				msg += " supplied to %s" % (hint)
			msg += ": "
			msg += " / ".join(errors)
			if len(self._required_arguments) > 0:
				msg += " -- required are: %s" % (", ".join(sorted(self._required_arguments)))
			if len(self._optional_arguments) > 0:
				msg += " -- allowed are: %s" % (", ".join(sorted(self._optional_arguments)))
			raise InvalidInputException(msg)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2020-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.estimate import JudgementCode, Compatibility
from x509sak.estimate.Judgement import SecurityJudgements, SecurityJudgement
from x509sak.Tools import DictTools

class BaseValidationResult():
	def __init__(self, validator, subject):
		self._validator = validator
		self._subject = subject
		self._result = SecurityJudgements()

	def _get_message(self, issue, message):
		return "%s %s" % (self._validator.validation_subject, message)

	def _report(self, issue_name, message, **kwargs):
		issue = self._validator.get_issue(issue_name)
		if issue is None:
#			print("Issue with no handler: %s" % (issue_name))
			return
		full_message = self._get_message(issue, message)
		if kwargs is None:
			kwargs = { }
		if issue.judgement is not None:
			kwargs.update(issue.judgement.kwargs)
		if ("standard" in kwargs) and ("compatibility" not in kwargs):
			kwargs["compatibility"] = Compatibility.STANDARDS_DEVIATION
		judgement = SecurityJudgement(issue.code, full_message, **kwargs)
		self._result += judgement

	def _delegate(self, subordinate_validator, subordinate_subject):
		validation_result = subordinate_validator.validate(subordinate_subject)
		self._result += validation_result

	def _validate(self):
		raise NotImplementedError(__class__.__name__)

	def run(self):
		self._validate()
		return self._result

class ValidationJudgement():
	def __init__(self, standard = None, info_payload = None):
		self._standard = standard
		self._info_payload = info_payload

	@property
	def standard(self):
		return self._standard

	@property
	def info_payload(self):
		return self._info_payload

	@property
	def kwargs(self):
		kwargs = { }
		if self.standard is not None:
			kwargs["standard"] = self.standard
		if self.info_payload is not None:
			kwargs["info_payload"] = self.info_payload
		return kwargs

	def __repr__(self):
		return "ValidationJudgement<std = %s, payload = %s>" % (self.standard, self.info_payload)

class ValidationIssue():
	def __init__(self, code = None, judgement = None):
		assert((judgement is None) or isinstance(judgement, ValidationJudgement))
		self._code = code
		self._judgement = judgement

	@property
	def code(self):
		return self._code

	@property
	def judgement(self):
		return self._judgement

	def __repr__(self):
		return "Issue(%s / %s)" % (self._code.name, str(self.judgement))

class BaseValidator():
	_ValidationResultClass = None

	def __init__(self, validation_subject, recognized_issues):
		assert(isinstance(recognized_issues, dict))
		assert(all(isinstance(value, ValidationIssue) for value in recognized_issues.values()))
		self._validation_subject = validation_subject
		self._recognized_issues = recognized_issues

	@property
	def validation_subject(self):
		return self._validation_subject

	@classmethod
	def create_inherited(cls, root_point_name, specific_judgements = None, **kwargs):
		if specific_judgements is None:
			specific_judgements = { }
		else:
			specific_judgements = DictTools.expand_multikeys(specific_judgements)

		invalid_keys = set(specific_judgements.keys()) - set(JudgementCode.inheritance[root_point_name].keys())
		if len(invalid_keys) > 0:
			raise Exception("Invalid specific_judgements passed: %s" % (", ".join(sorted(invalid_keys))))

		recognized_issues = { name: ValidationIssue(code = code, judgement = specific_judgements.get(name)) for (name, code) in JudgementCode.inheritance[root_point_name].items() }
		return cls(recognized_issues = recognized_issues, **kwargs)

	def get_issue(self, issue_name):
		return self._recognized_issues.get(issue_name)

	def validate(self, subject):
		validation_result = self._ValidationResultClass(self, subject)
		return validation_result.run()



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2020-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import json
from x509sak.tests import BaseTest
from x509sak.estimate import JudgementCode
from x509sak.X509Certificate import X509Certificate
from x509sak.tests import ResourceFileLoader
from x509sak.CertificateAnalyzer import CertificateAnalyzer
from x509sak.Tools import FileLockTools
from x509sak.Exceptions import UnexpectedFileContentException

class BaseAnalyzerTest(BaseTest):
	def _update_stats_file(self, certname, parent_certname, encountered_codes, checked_codes):
		stats_filename = ".examinecert_stats.json"
		with FileLockTools.lock(stats_filename):
			try:
				with open(stats_filename) as f:
					stats = json.load(f)
			except (FileNotFoundError, json.JSONDecodeError):
				stats = { }

			if "encountered_codes" not in stats:
				stats["encountered_codes"] = { }
			for encountered_code in encountered_codes:
				if encountered_code not in stats["encountered_codes"]:
					stats["encountered_codes"][encountered_code] = [ certname, parent_certname ]

			if "checked_codes" not in stats:
				stats["checked_codes"] = { }
			for checked_code in checked_codes:
				if checked_code not in stats["checked_codes"]:
					stats["checked_codes"][checked_code] = [ certname, parent_certname ]

			with open(stats_filename, "w") as f:
				json.dump(stats, f)

	def _test_examine_x509test_resultcode(self, certname, expect_present = None, expect_absent = None, parent_certname = None, fast_rsa = True, host_check = None, include_raw = False, purpose = None, expect_parse_failure = False):
		if expect_present is None:
			expect_present = tuple()
		if not isinstance(expect_present, (list, tuple)):
			expect_present = (expect_present, )

		if expect_absent is None:
			expect_absent = tuple()
		if not isinstance(expect_absent, (list, tuple)):
			expect_absent = (expect_absent, )

		# Plausibilize we're not chasing non-existing judgement codes -- don't
		# check the empty string because we often use that for debugging and
		# it's *obviously* wrong.

		if expect_present != ("", ):
			self.assertTrue(all(getattr(JudgementCode, codename, None) is not None for codename in expect_present))
		if expect_absent != ("", ):
			self.assertTrue(all(getattr(JudgementCode, codename, None) is not None for codename in expect_absent))

		if expect_parse_failure:
			with self.assertRaises(UnexpectedFileContentException):
				X509Certificate.from_pem_data(ResourceFileLoader.load_data(certname))
			return

		certificates = X509Certificate.from_pem_data(ResourceFileLoader.load_data(certname))
		crt_sources = [ CertificateAnalyzer.CertSource(crts = certificates, source = "internal", source_type = "pemcert") ]
		if parent_certname is not None:
			ca_certificate = CertificateAnalyzer.CertSource(crts = X509Certificate.from_pem_data(ResourceFileLoader.load_data(parent_certname)), source = "internal", source_type = "pemcert")
		else:
			ca_certificate = None

		analysis_params = {
			"fast_rsa":				fast_rsa,
			"include_raw_data":		include_raw,
		}
		if host_check is not None:
			analysis_params.update({
				"entity_name":	host_check,
				"purposes":		[ "tls-server" ],
			})
		elif purpose is not None:
			analysis_params["purposes"] = [ purpose ]
		cert_analyzer = CertificateAnalyzer(**analysis_params)
		analyses = cert_analyzer.analyze(crt_sources, ca_certificate)

		encountered_codes = CertificateAnalyzer.extract_codes_from_json(analyses)

		# If we're in debugging mode, update the consolidated JSON stat file
		if self._produce_statistics:
			self._update_stats_file(certname = certname, parent_certname = parent_certname, encountered_codes = encountered_codes, checked_codes = expect_present)
		for code in expect_present:
			self.assertIn(code, encountered_codes)
		for code in expect_absent:
			self.assertNotIn(code, encountered_codes)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import math
from x509sak.NumberTheory import NumberTheory
from x509sak.estimate.BaseEstimator import BaseEstimator
from x509sak.estimate import JudgementCode, Commonness, Compatibility
from x509sak.estimate.Judgement import SecurityJudgement, SecurityJudgements, LiteratureReference

@BaseEstimator.register
class DSASecurityEstimator(BaseEstimator):
	_ALG_NAME = "dsa"

	"""DSA Parameters:
	p: Prime modulus of bitlength L
	q: Prime divisor of (p - 1) of bitlength N
	g: Generator of order q in GF(p); 1 < g < p
	y: pubkey, y = g^x mod p; x is the private key
	"""

	_TYPICAL_L_N_VALUES = {
		1024:	(160, ),
		2048:	(224, 256),
		3072:	(256, )
	}

	def analyze(self, pubkey):
		judgements = SecurityJudgements()

		L = pubkey.p.bit_length()
		N = pubkey.q.bit_length()

		if not NumberTheory.is_probable_prime(pubkey.p):
			standard = LiteratureReference(quote = "p: a prime modulus", sect = "4.1", author = "National Institute of Standards and Technology", title = "FIPS PUB 186-4: Digital Signature Standard (DSS)", year = 2013, month = 7, doi = "10.6028/NIST.FIPS.186-4")
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_DSA_Parameters_P_NotPrime, "DSA parameter p is not prime.", commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION, bits = 0, standard = standard)

		if not NumberTheory.is_probable_prime(pubkey.q):
			standard = LiteratureReference(quote = "q: a prime divisor of (p - 1)", sect = "4.1", author = "National Institute of Standards and Technology", title = "FIPS PUB 186-4: Digital Signature Standard (DSS)", year = 2013, month = 7, doi = "10.6028/NIST.FIPS.186-4")
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_DSA_Parameters_Q_NotPrime, "DSA parameter q is not prime.", commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION, bits = 0, standard = standard)

		if ((pubkey.p - 1) % pubkey.q) != 0:
			standard = LiteratureReference(quote = "q: a prime divisor of (p - 1)", sect = "4.1", author = "National Institute of Standards and Technology", title = "FIPS PUB 186-4: Digital Signature Standard (DSS)", year = 2013, month = 7, doi = "10.6028/NIST.FIPS.186-4")
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_DSA_Parameters_Q_NoDivisorOfP1, "DSA parameter q is not a divisor of (p - 1).", commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION, bits = 0, standard = standard)

		if pow(pubkey.g, pubkey.q, pubkey.p) != 1:
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_DSA_Parameters_G_Invalid, "DSA parameter g is not valid. In particular, g^q mod p != 1.", commonness = Commonness.HIGHLY_UNUSUAL, bits = 0)

		if (pubkey.g <= 1) or (pubkey.g >= pubkey.p):
			standard = LiteratureReference(quote = "g: a generator of a subgroup of order q in the multiplicative group of GF(p), such that 1 < g < p", sect = "4.1", author = "National Institute of Standards and Technology", title = "FIPS PUB 186-4: Digital Signature Standard (DSS)", year = 2013, month = 7, doi = "10.6028/NIST.FIPS.186-4")
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_DSA_Parameters_G_InvalidRange, "DSA parameter g is not inside the valid range (1 < g < p).", commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION, bits = 0, standard = standard)

		hweight_analysis = NumberTheory.hamming_weight_analysis(pubkey.p)
		if not hweight_analysis.plausibly_random:
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_DSA_Parameters_P_BitBiasPresent, "Hamming weight of DSA prime p is %d at bitlength %d, but expected a weight between %d and %d when randomly chosen; this is likely not coincidential." % (hweight_analysis.hweight, hweight_analysis.bitlen, hweight_analysis.rnd_min_hweight, hweight_analysis.rnd_max_hweight), commonness = Commonness.HIGHLY_UNUSUAL)

		hweight_analysis = NumberTheory.hamming_weight_analysis(pubkey.q)
		if not hweight_analysis.plausibly_random:
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_DSA_Parameters_Q_BitBiasPresent, "Hamming weight of DSA prime q is %d at bitlength %d, but expected a weight between %d and %d when randomly chosen; this is likely not coincidential." % (hweight_analysis.hweight, hweight_analysis.bitlen, hweight_analysis.rnd_min_hweight, hweight_analysis.rnd_max_hweight), commonness = Commonness.HIGHLY_UNUSUAL)

		if (L in self._TYPICAL_L_N_VALUES) and (N in self._TYPICAL_L_N_VALUES[L]):
			# Typical
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_DSA_L_N_Common, "DSA parameter values L/N (%d/%d) are common." % (L, N), commonness = Commonness.COMMON)
		else:
			# Non-typical
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_DSA_L_N_Uncommon, "DSA parameter values L/N (%d/%d) are uncommon." % (L, N), commonness = Commonness.UNUSUAL)

		L_strength_bits = NumberTheory.asymtotic_complexity_gnfs_bits(pubkey.p)
		N_strength_bits = math.floor(N / 2)
		bits_security = min(L_strength_bits, N_strength_bits)
		judgements += self.algorithm("bits").analyze(JudgementCode.X509Cert_PublicKey_DSA_L_N, bits_security)

		result = {
			"cryptosystem":	"dsa",
			"specific": {
				"L":	L,
				"N":	N,
			},
			"security": judgements,
		}

		if self._analysis_options.include_raw_data:
			result["specific"]["p"] = {
				"value": pubkey.p,
			}
			result["specific"]["q"] = {
				"value": pubkey.q,
			}
			result["specific"]["g"] = {
				"value": pubkey.g,
			}
			result["specific"]["pubkey"] = {
				"value": pubkey.pubkey,
			}
		return result



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2020-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

class RXBuffer():
	def __init__(self):
		self._data = bytearray()

	def get_all(self):
		return self.get(len(self))

	def read_from(self, source, until):
		while True:
			if until(self):
				break
			data = source(self)
			if len(data) > 0:
				self += data

	def have_bytes(self, count):
		return len(self._data) >= count

	def have_line(self):
		return self._data.find(b"\n") != -1

	def get(self, length):
		if len(self._data) < length:
			return None
		else:
			chunk = self._data[:length]
			self._data = self._data[length:]
			return chunk

	def getline(self):
		index = self._data.find(b"\n")
		if index == -1:
			return None
		else:
			return self.get(index + 1)

	def __iadd__(self, data):
		self._data += data
		return self

	def __len__(self):
		return len(self._data)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2019 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import enum
import urllib.parse
from x509sak.KwargsChecker import KwargsChecker
from x509sak.Exceptions import InvalidInputException, LazyDeveloperException, UnexpectedFileContentException
from x509sak.FriendlyArgumentParser import baseint
from x509sak.RSAPrivateKey import RSAPrivateKey
from x509sak.ECPrivateKey import ECPrivateKey
from x509sak.EDPrivateKey import EDPrivateKey

class PrivateKeyStorageForm(enum.IntEnum):
	PEM_FILE = 1
	DER_FILE = 2
	HARDWARE_TOKEN = 3

class PrivateKeyStorage():
	_PRIVATE_KEY_CLASSES = [ RSAPrivateKey, ECPrivateKey, EDPrivateKey ]

	_PARAMETER_CONSTRAINTS = {
		PrivateKeyStorageForm.PEM_FILE:			KwargsChecker(required_arguments = set([ "filename" ]), optional_arguments = set([ "search_path" ])),
		PrivateKeyStorageForm.DER_FILE:			KwargsChecker(required_arguments = set([ "filename" ]), optional_arguments = set([ "search_path" ])),
		PrivateKeyStorageForm.HARDWARE_TOKEN:	KwargsChecker(required_arguments = set([ "pkcs11uri", "so_search_path" ]), optional_arguments = set([ "dynamic_so", "module_so" ])),
	}

	def __init__(self, storage_form, **kwargs):
		assert(isinstance(storage_form, PrivateKeyStorageForm))
		self._storage_form = storage_form
		self._PARAMETER_CONSTRAINTS[storage_form].check(kwargs, hint = "PrivateKeyStorage using %s" % (storage_form.name))
		self._parameters = kwargs
		if "search_path" not in self._parameters:
			self._parameters["search_path"] = ""
		if self._storage_form == PrivateKeyStorageForm.HARDWARE_TOKEN:
			self._verify_pkcs11_uri()

	def _verify_pkcs11_uri(self):
		uri = self._parameters["pkcs11uri"]
		if uri.startswith("pkcs11:"):
			# Literal PKCS#11 URI, leave as-is.
			pass
		elif uri.startswith("label="):
			# Replace with encoded version
			label = uri[6 : ]
			self._parameters["pkcs11uri"] = "pkcs11:object=%s;type=private" % (urllib.parse.quote(label))
		elif uri.startswith("id="):
			# Replace with encoded version
			key_id_str = uri[3 : ]
			try:
				key_id = baseint(key_id_str)
				key_bytes = key_id.to_bytes(length = 16, byteorder = "big").lstrip(bytes(1))
			except (ValueError, OverflowError) as e:
				raise InvalidInputException("Key ID '%s' is not a valid hex value or is too large: %s" % (key_id_str, e.__class__.__name__))
			key_id_quoted = "".join("%%%02x" % (c) for c in key_bytes)
			self._parameters["pkcs11uri"] = "pkcs11:id=%s;type=private" % (key_id_quoted)
		else:
			raise InvalidInputException("For hardware keys, you need to either give a RFC7512-compliant pkcs11-scheme URI (starts with 'pkcs11:'), a key label in the form 'label=foobar' or a key id in the hex form like 'id=0xabc123' or in decimal form like 'id=11256099'. The supplied value '%s' is neither." % (uri))

	def update(self, key, value):
		self._PARAMETER_CONSTRAINTS[self._storage_form].check_single(key)
		self._parameters[key] = value

	@property
	def storage_form(self):
		return self._storage_form

	@property
	def is_file_based(self):
		return self.storage_form in [ PrivateKeyStorageForm.PEM_FILE, PrivateKeyStorageForm.DER_FILE ]

	@property
	def filename(self):
		assert(self.storage_form in [ PrivateKeyStorageForm.PEM_FILE, PrivateKeyStorageForm.DER_FILE ])
		return self._parameters["filename"]

	@property
	def full_filename(self):
		assert(self.storage_form in [ PrivateKeyStorageForm.PEM_FILE, PrivateKeyStorageForm.DER_FILE ])
		return self._parameters["search_path"] + self._parameters["filename"]

	@property
	def pkcs11uri(self):
		assert(self.storage_form == PrivateKeyStorageForm.HARDWARE_TOKEN)
		return self._parameters["pkcs11uri"]

	@property
	def so_search_path(self):
		assert(self.storage_form == PrivateKeyStorageForm.HARDWARE_TOKEN)
		return self._parameters["so_search_path"]

	@property
	def dynamic_so(self):
		assert(self.storage_form == PrivateKeyStorageForm.HARDWARE_TOKEN)
		return self._parameters.get("dynamic_so", "libpkcs11.so")

	@property
	def module_so(self):
		assert(self.storage_form == PrivateKeyStorageForm.HARDWARE_TOKEN)
		return self._parameters.get("module_so", "opensc-pkcs11.so")

	def to_dict(self):
		return {
			"storage_form":		self.storage_form.name,
			"parameters":		{ key: value for (key, value) in self._parameters.items() if key not in [ "search_path" ] },
		}

	@classmethod
	def from_dict(cls, serialized_dict, **kwargs):
		storage_form = getattr(PrivateKeyStorageForm, serialized_dict["storage_form"])
		parameters = serialized_dict["parameters"]
		parameters.update(kwargs)
		return cls(storage_form = storage_form, **parameters)

	@classmethod
	def from_str(cls, key_type, key_value):
		if key_type == "pem":
			return cls(PrivateKeyStorageForm.PEM_FILE, filename = key_value)
		elif key_type == "der":
			return cls(PrivateKeyStorageForm.DER_FILE, filename = key_value)
		elif key_type == "hw":
			return cls(PrivateKeyStorageForm.HARDWARE_TOKEN, pkcs11uri = key_value)
		else:
			raise LazyDeveloperException(NotImplemented, key_type)

	def load_private_key(self):
		assert(self.is_file_based)
		for handler_class in self._PRIVATE_KEY_CLASSES:
			try:
				if self.storage_form == PrivateKeyStorageForm.PEM_FILE:
					privkey = handler_class.read_pemfile(self.full_filename)[0]
				elif self.storage_form == PrivateKeyStorageForm.DER_FILE:
					privkey = handler_class.read_derfile(self.full_filename)
				else:
					raise LazyDeveloperException(NotImplemented, self.storage_form)
			except UnexpectedFileContentException:
				continue
			return privkey
		raise UnexpectedFileContentException("Could not load key material from %s." % (self.full_filename))

	def __str__(self):
		if self.storage_form in [ PrivateKeyStorageForm.PEM_FILE, PrivateKeyStorageForm.DER_FILE ]:
			return "PrivateKeyStorage<%s: %s>" % (self.storage_form.name, self.filename)
		elif self.storage_form == PrivateKeyStorageForm.HARDWARE_TOKEN:
			return "PrivateKeyStorage<%s %s>" % (self.storage_form.name, self.pkcs11uri)
		else:
			return "PrivateKeyStorage<%s: %s>" % (self.storage_form.name, str(self._parameters))



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2020-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import sys
import contextlib
import itertools
import subprocess
import collections
from x509sak.BaseAction import BaseAction
from x509sak.X509Certificate import X509Certificate
from x509sak.OID import OID
from x509sak.estimate import JudgementCode
from x509sak.certgen.CertGenerator import CertGenerator
from x509sak.Tools import PEMDataTools

class ActionTestcaseGen(BaseAction):
	_EXPECT_PRESENT_CODEPOINT_OID = OID.from_str("1.13.99.127.41")
	_TCDefinition = collections.namedtuple("TCDefinition", [ "full_name", "filename", "expect_present" ])

	def __init__(self, cmdname, args):
		BaseAction.__init__(self, cmdname, args)

		sys.setrecursionlimit(10000)
		self._cg = CertGenerator.instantiate(self._args.tcname + ".ader")
		if self._args.list_parameters:
			for (name, values) in sorted(self._cg.parameters):
				print("%-30s %s" % (name, ", ".join(sorted(values))))
			sys.exit(0)

		template_parameters = { }
		for parameter in self._args.parameter:
			(key, value) = parameter.split("=", maxsplit = 1)
			if value == "*":
				template_parameters[key] = self._cg.get_choices(key)
			else:
				template_parameters[key] = [ value ]

		for (name, values) in self._cg.parameters:
			if name not in template_parameters:
				template_parameters[name] = values

		with contextlib.suppress(FileExistsError):
			os.makedirs(self._args.output_dir)

		template_parameters = list(template_parameters.items())
		keys = [ key for (key, values) in template_parameters ]
		values = [ values for (key, values) in template_parameters ]
		self._tcs = [ ]
		for concrete_values in itertools.product(*values):
			concrete_values = dict(zip(keys, concrete_values))
			try:
				self._render(concrete_values)
			except subprocess.CalledProcessError as e:
				print("Failed: %s (%s)" % (str(concrete_values), str(e)))

		self._tcs.sort()
		self._print_tcs()

	@staticmethod
	def testcase_codepoint_string(codepoints):
		if len(codepoints) == 1:
			return "\"%s\"" % (codepoints[0])
		else:
			return "[ " + ", ".join("\"%s\"" % (codepoint) for codepoint in codepoints)  + " ]"

	def _print_tcs(self):
		with open(self._args.output_dir + "/tcs.txt", "w") as f:
			print("# " + ("=" * 70) + " Begin of %s " % (self._args.tcname) + ("=" * 70), file = f)
			for tc in self._tcs:
				print("\tdef %s(self):" % (tc.full_name), file = f)
				print("\t\tself._test_examine_x509test_resultcode(\"%s\", expect_present = %s)" % (tc.filename, self.testcase_codepoint_string(tc.expect_present)), file = f)
				print(file = f)
			print("# " + ("=" * 70) + " End of %s " % (self._args.tcname) + ("=" * 70), file = f)

	def _extract_codepoints(self, filename_prefix, pem_file):
		x509cert = X509Certificate.from_pem_data(pem_file)[0]
		present_codepoints = [ rdn.get_value(self._EXPECT_PRESENT_CODEPOINT_OID).printable_value for rdn in x509cert.subject.get_all(self._EXPECT_PRESENT_CODEPOINT_OID) ]
		present_codepoints += [ rdn.get_value(self._EXPECT_PRESENT_CODEPOINT_OID).printable_value for rdn in x509cert.issuer.get_all(self._EXPECT_PRESENT_CODEPOINT_OID) ]

		codepoints = [ ]
		for codepoint_name in present_codepoints:
			if "#" in codepoint_name:
				continue
			try:
				getattr(JudgementCode, codepoint_name)
			except AttributeError:
				print("No such codepoint: %s (in %s)" % (codepoint_name, filename_prefix))
				continue
			codepoints.append(codepoint_name)
		return codepoints

	def _render(self, concrete_values):
		if self._args.verbose >= 1:
			print(concrete_values)
		(filename_prefix, ascii_der_source) = self._cg.render(concrete_values)

		if self._args.no_pem:
			self._write_raw_ader_file(filename_prefix, ascii_der_source)
		else:
			self._write_pem_file(filename_prefix, ascii_der_source)

	def _write_raw_ader_file(self, filename_prefix, ascii_der_source):
		outfile = self._args.output_dir + "/" + basename + ".ader"
		with open(outfile, "w") as f:
			f.write(ascii_der_source)
		print(ascii_der_source)

	def _write_pem_file(self, filename_prefix, ascii_der_source):
		der_data = subprocess.check_output([ "ascii2der" ], input = ascii_der_source.encode())
		try:
			pem_cert = subprocess.check_output([ "openssl", "x509", "-inform", "der", "-text" ], input = der_data)
			pem_cert = pem_cert.decode()
			pem_cert = [ line.rstrip("\t ") for line in pem_cert.split("\n") ]
			pem_cert = "\n".join(pem_cert)
		except subprocess.CalledProcessError:
			# OpenSSL cannot encode the certificate
			pem_cert = PEMDataTools.data2pem(der_data, marker = "CERTIFICATE")

		codepoints = self._extract_codepoints(filename_prefix, pem_cert)
		if len(codepoints) == 0:
			return

		outfile = self._args.output_dir + "/" + filename_prefix + ".pem"
		with open(outfile, "w") as f:
			f.write(pem_cert)

		tc_filename = "certs/generated/%s/%s.pem" % (self._args.tcname, filename_prefix)
		tc = self._TCDefinition(full_name = "test_generated_%s_%s" % (self._args.tcname, filename_prefix), filename = tc_filename, expect_present = codepoints)
		self._tcs.append(tc)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import base64
import pyasn1.codec.der.decoder
from pyasn1_modules import rfc2459
from x509sak.tests import BaseTest
from x509sak.Exceptions import InvalidInputException
from x509sak.DistinguishedName import RelativeDistinguishedName, DistinguishedName

class DistinguishedNameTests(BaseTest):
	def test_simple_rdn_1(self):
		rdn1 = RelativeDistinguishedName.create("CN", "Rick Astley")
		rdn2 = RelativeDistinguishedName.create("CN", "Rick Astley")
		rdn3 = RelativeDistinguishedName.create("CN", "Darude")
		self.assertEqual(rdn1, rdn2)
		self.assertNotEqual(rdn1, rdn3)
		self.assertEqual(rdn1.rfc2253_str, "CN=Rick Astley")

	def test_simple_mvrdn_1(self):
		# Order of elements in RDN does not matter
		rdn = RelativeDistinguishedName.create("CN", "Rick Astley", "OU", "Infinite Persistence Systems Inc.")
		self.assertIn(rdn.rfc2253_str, [ "CN=Rick Astley+OU=Infinite Persistence Systems Inc.", "OU=Infinite Persistence Systems Inc.+CN=Rick Astley" ])

	def test_simple_mvrdn_2(self):
		rdn = RelativeDistinguishedName.create("CN", "Rick Astley", "CN", "Foo Bar")
		self.assertIn(rdn.rfc2253_str, [ "CN=Rick Astley+CN=Foo Bar", "CN=Foo Bar+CN=Rick Astley" ])

	def test_simple_dn_1(self):
		elements = [ RelativeDistinguishedName.create("CN", "Rick Astley") ]
		dn = DistinguishedName(elements)
		self.assertEqual(dn.rfc2253_str, "CN=Rick Astley")

	def test_simple_dn_2(self):
		elements = [ RelativeDistinguishedName.create("CN", "Rick Astley"), RelativeDistinguishedName.create("OU", "Infinite Persistence Systems Inc.") ]
		dn1 = DistinguishedName(elements)
		self.assertEqual(dn1.rfc2253_str, "OU=Infinite Persistence Systems Inc.,CN=Rick Astley")

		elements = [ RelativeDistinguishedName.create("OU", "Infinite Persistence Systems Inc."), RelativeDistinguishedName.create("CN", "Rick Astley") ]
		dn2 = DistinguishedName(elements)
		self.assertEqual(dn2.rfc2253_str, "CN=Rick Astley,OU=Infinite Persistence Systems Inc.")

		self.assertNotEqual(dn1, dn2)

	def test_simple_dn_3(self):
		elements = [ RelativeDistinguishedName.create("CN", "Rick Astley"), RelativeDistinguishedName.create("CN", "Foo Bar"), RelativeDistinguishedName.create("OU", "OrgUnit"), RelativeDistinguishedName.create("CN", "Bar Foo") ]
		dn = DistinguishedName(elements)
		self.assertEqual(dn.rfc2253_str, "CN=Bar Foo,OU=OrgUnit,CN=Foo Bar,CN=Rick Astley")

	def test_json_DNs(self):
		for testcase in self._load_json("misc/dn_tests.json"):
			rdn_sequence_derdata = base64.b64decode(testcase["rdn_sequence"].encode())
			(rdn_sequence_asn1, _) = pyasn1.codec.der.decoder.decode(rdn_sequence_derdata, asn1Spec = rfc2459.Name())
			dn = DistinguishedName.from_asn1(rdn_sequence_asn1)
			self.assertEqual(dn.rfc2253_str, testcase["rfc2253"])

	def test_decoded_asn1_different_types(self):
		(utf8, _) = pyasn1.codec.der.decoder.decode(bytes.fromhex("30 16 31 14 30 12 06 03 55 04 03 0c 0b 43 6f 6d 6d 6f 6e 20 4e 61 6d 65"), asn1Spec = rfc2459.Name())
		(ia5, _) = pyasn1.codec.der.decoder.decode(bytes.fromhex("30 16 31 14 30 12 06 03 55 04 03 16 0b 43 6f 6d 6d 6f 6e 20 4e 61 6d 65"), asn1Spec = rfc2459.Name())
		utf8dn = DistinguishedName.from_asn1(utf8)
		ia5dn = DistinguishedName.from_asn1(ia5)
		self.assertEqual(utf8dn, ia5dn)

	def test_decoded_asn1_different_types_umlaut(self):
		(utf8, _) = pyasn1.codec.der.decoder.decode(bytes.fromhex("30 17 31 15 30 13 06 03 55 04 03 0c 0c 43 c3 b6 6d 6d 6f 6e 20 4e 61 6d 65"), asn1Spec = rfc2459.Name())
		(bmp, _) = pyasn1.codec.der.decoder.decode(bytes.fromhex("30 21 31 1f 30 1d 06 03 55 04 03 1e 16 00 43 00 f6 00 6d 00 6d 00 6f 00 6e 00 20 00 4e 00 61 00 6d 00 65"), asn1Spec = rfc2459.Name())
		utf8dn = DistinguishedName.from_asn1(utf8)
		bmpdn = DistinguishedName.from_asn1(bmp)
		self.assertEqual(utf8dn, bmpdn)

	def test_equality_rdn(self):
		rdn1 = RelativeDistinguishedName.create("CN", "Foo", "O", "Bar")
		rdn2 = RelativeDistinguishedName.create("O", "Bar", "CN", "Foo")
		self.assertEqual(rdn1, rdn2)

	def test_equality_dn(self):
		dn1 = DistinguishedName([ RelativeDistinguishedName.create("CN", "Foo"), RelativeDistinguishedName.create("CN", "Bar") ])
		dn2 = DistinguishedName([ RelativeDistinguishedName.create("CN", "Bar"), RelativeDistinguishedName.create("CN", "Foo") ])
		dn3 = DistinguishedName([ RelativeDistinguishedName.create("CN", "Foo"), RelativeDistinguishedName.create("CN", "Bar") ])
		self.assertNotEqual(dn1, dn2)
		self.assertEqual(dn1, dn3)

	def test_string_parse(self):
		dn = DistinguishedName.from_rfc2253_str("cn=foo,ou=bar,o=koo")
		self.assertEqual(dn, DistinguishedName([ RelativeDistinguishedName.create("CN", "foo"), RelativeDistinguishedName.create("OU", "bar"), RelativeDistinguishedName.create("O", "koo") ]))

		dn = DistinguishedName.from_rfc2253_str("cn=foo,ou=bar,o=koo+ou=moo")
		self.assertEqual(dn, DistinguishedName([ RelativeDistinguishedName.create("CN", "foo"), RelativeDistinguishedName.create("OU", "bar"), RelativeDistinguishedName.create("O", "koo", "OU", "moo") ]))

		dn = DistinguishedName.from_rfc2253_str("")
		self.assertEqual(dn, DistinguishedName([ ]))

	def test_string_parse_fail(self):
		with self.assertRaises(InvalidInputException):
			DistinguishedName.from_rfc2253_str("/")

		with self.assertRaises(InvalidInputException):
			DistinguishedName.from_rfc2253_str("CN,OU=bar")

		with self.assertRaises(InvalidInputException):
			DistinguishedName.from_rfc2253_str("MUH=KUH")



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from pyasn1.type import tag, namedtype, univ
from x509sak.PEMDERObject import PEMDERObject
from x509sak.Tools import ASN1Tools, ECCTools
from x509sak.OID import OID
from x509sak.Exceptions import InvalidInputException
from x509sak.AlgorithmDB import Cryptosystems
from x509sak.PublicKey import PublicKey
from x509sak.CurveDB import CurveDB

class _ECPrivateKey(univ.Sequence):
	"""Minimalistic RFC5915 implementation."""

	componentType = namedtype.NamedTypes(
		namedtype.NamedType("version", univ.Integer()),
		namedtype.NamedType("privateKey", univ.OctetString()),
		namedtype.NamedType("parameters", univ.ObjectIdentifier().subtype(explicitTag = tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 0))),
		namedtype.NamedType("publicKey", univ.BitString().subtype(explicitTag = tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 1))),
	)

class ECPrivateKey(PEMDERObject):
	_PEM_MARKER = "EC PRIVATE KEY"
	_ASN1_MODEL = _ECPrivateKey

	def _post_decode_hook(self):
		if self.asn1["parameters"] is None:
			raise InvalidInputException("ECC private key does not contain curve OID. Cannot proceed.")
		if self.asn1["publicKey"] is None:
			raise InvalidInputException("ECC private key does not contain public key. Cannot proceed.")

		curve_oid = OID.from_asn1(self.asn1["parameters"])
		self._curve = CurveDB().instantiate(oid = curve_oid)

		self._d = int.from_bytes(self.asn1["privateKey"], byteorder = "big")
		(self._x, self._y) = ECCTools.decode_enc_pubkey(ASN1Tools.bitstring2bytes(self.asn1["publicKey"]))

	@property
	def cryptosystem(self):
		return Cryptosystems.ECC_ECDSA

	@property
	def pubkey(self):
		return PublicKey.create(cryptosystem = self.cryptosystem, parameters = { "curve": self.curve, "x": self.x, "y": self.y })

	@property
	def curve(self):
		return self._curve

	@property
	def d(self):
		return self._d

	@property
	def x(self):
		return self._x

	@property
	def y(self):
		return self._y

	def __str__(self):
		return "ECPrivateKey<%s>" % (self.curve)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2020-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import datetime
import collections
from x509sak.estimate import SecurityEstimator, AnalysisOptions
from x509sak.Tools import JSONTools

class CertificateAnalyzer():
	CertSource = collections.namedtuple("CertSource", [ "source", "source_type", "crts" ])

	def __init__(self, entity_name = None, fast_rsa = False, purposes = None, include_raw_data = False):
		self._entity_name = entity_name
		self._fast_rsa = fast_rsa
		self._purposes = purposes if (purposes is not None) else [ ]
		self._purposes = list(sorted(AnalysisOptions.CertificatePurpose(purpose) for purpose in set(self._purposes)))
		self._include_raw_data = include_raw_data

	def analyze(self, crt_sources, root_certificate):
		assert(isinstance(crt_source, self.CertSource) for crt_source in crt_sources)
		assert(isinstance(root_certificate, (type(None), self.CertSource)))
		utcnow = datetime.datetime.utcnow()
		analyses = {
			"timestamp_utc":	utcnow,
			"data":				[ ],
		}
		for crt_source in crt_sources:
			for (crtno, crt) in enumerate(crt_source.crts, 1):
				analysis_options = {
					"rsa_testing":			AnalysisOptions.RSATesting.Fast if self._fast_rsa else AnalysisOptions.RSATesting.Full,
					"include_raw_data":		self._include_raw_data,
					"purposes":				self._purposes,
					"fqdn":					self._entity_name,
					"utcnow":				utcnow,
				}
				if root_certificate is not None:
					root_crt = root_certificate.crts[0]
				else:
					root_crt = None
				analysis_options = AnalysisOptions(**analysis_options)
				analysis = SecurityEstimator.handler("certificate")(analysis_options = analysis_options).analyze(crt, root_crt)
				analysis = JSONTools.translate(analysis)
				analysis["source"] = {
					"name":			crt_source.source,
					"srctype":		crt_source.source_type,
					"cert_no":		crtno,
					"certs_total":	len(crt_source.crts),
				}
				if root_certificate is not None:
					analysis["root_source"] = {
						"name":			root_certificate.source,
						"srctype":		root_certificate.source_type,
					}
				analyses["data"].append(analysis)
		return analyses

	@classmethod
	def extract_codes_from_json(cls, data):
		def recurse_through_data(data, result):
			if result is None:
				result = set()
			if isinstance(data, list):
				for item in data:
					recurse_through_data(item, result)
			elif isinstance(data, dict):
				if "code" in data:
					result.add(data["code"])
				for (key, value) in data.items():
					recurse_through_data(value, result)

		result = set()
		recurse_through_data(data, result)
		return result
